import numpy as np
from typing import *
from pathlib import Path
import os

from .. import custom_paths
from .. import utils

class ExperimentResults:
    def __init__(self, results_dict: dict, exp_name: str):
        # usually, results_dict[alg_name][task_name][split_idx] = something
        self.results_dict = results_dict
        self.exp_name = exp_name
        self.alg_names = list(self.results_dict.keys())
        # self.alg_names.sort()
        self.task_names = list(set.union(*[set(alg_results.keys()) for alg_results in self.results_dict.values()]))
        self.task_names.sort()

    def map_single_split(self, f: Callable):
        # can be used e.g. to compute average log metric results
        return ExperimentResults({alg_name: {task_name: [f(split) for split in split_results]
                                             for task_name, split_results in task_results.items()}
                                  for alg_name, task_results in self.results_dict.items()}, exp_name=self.exp_name)

    def get_learning_curves(self, key: str) -> 'ExperimentResults':
        return ExperimentResults({alg_name: {task_name: np.mean(np.log([split['errors'][key] for split in split_results]), axis=0)
                                             for task_name, split_results in task_results.items() if int(task_name.split('_')[-1].split('x')[0]) == 256}
                                  for alg_name, task_results in self.results_dict.items()}, exp_name=self.exp_name)

    def get_avg_al_stats(self, key: str) -> 'ExperimentResults':
        return self.map_single_split(
            lambda split_dict: [stat[key] for stat in split_dict['al_stats'] if key in stat])

    def get_avg_al_times(self, key: str) -> 'ExperimentResults':
        return self.map_single_split(
            lambda split_dict: [stat[key]['total'] for stat in split_dict['al_stats'] if key in stat])

    def get_avg_errors(self, key: str, use_log: bool = True) -> 'ExperimentResults':
        if use_log:
            return self.map_single_split(
                lambda split_dict: np.mean(np.log(split_dict['errors'][key][1:])))
        else:
            return self.map_single_split(
                lambda split_dict: np.mean(split_dict['errors'][key][1:]))

    def get_last_errors(self, key: str, use_log: bool = True) -> 'ExperimentResults':
        if use_log:
            return self.map_single_split(
                lambda split_dict: np.mean(np.log(split_dict['errors'][key][-1])))
        else:
            return self.map_single_split(
                lambda split_dict: np.mean(split_dict['errors'][key][-1]))

    def get_average_al_times(self) -> 'ExperimentResults':
        return self.map_single_split(
            lambda split_dict: np.mean([sr['kernel_time']['total'] + sr['selection_time']['total']
                                        for sr in split_dict['al_stats']]))

    def select_split(self, i: int) -> 'ExperimentResults':
        for alg_name, task_results in self.results_dict.items():
            for task_name, split_results in task_results.items():
                if len(split_results) <= i:
                    print(f'Invalid index for alg {alg_name} on task {task_name}')
        return ExperimentResults({alg_name: {task_name: split_results[i]
                                             for task_name, split_results in task_results.items()}
                                  for alg_name, task_results in self.results_dict.items()}, exp_name=self.exp_name)

    def filter_task_suffix(self, task_suffix: str) -> 'ExperimentResults':
        return ExperimentResults(
            {alg_name: {task_name: task_dict
                        for task_name, task_dict in alg_dict.items() if task_name.endswith(task_suffix)}
            for alg_name, alg_dict in self.results_dict.items()}, exp_name=self.exp_name)

    def filter_task_prefix(self, task_prefix: str) -> 'ExperimentResults':
        return ExperimentResults(
            {alg_name: {task_name: task_dict
                        for task_name, task_dict in alg_dict.items() if task_name.startswith(task_prefix)}
            for alg_name, alg_dict in self.results_dict.items()}, exp_name=self.exp_name)

    def filter_task_names(self, task_names: List[str]) -> 'ExperimentResults':
        return ExperimentResults(
            {alg_name: {task_name: task_dict
                        for task_name, task_dict in alg_dict.items() if task_name in task_names}
            for alg_name, alg_dict in self.results_dict.items()}, exp_name=self.exp_name)

    def filter_alg_names(self, alg_names: Iterable[str]) -> 'ExperimentResults':
        return ExperimentResults({alg_name: self.results_dict[alg_name] for alg_name in alg_names
                                  if alg_name in self.results_dict}, exp_name=self.exp_name)

    def filter_common_algs(self) -> 'ExperimentResults':
        common_alg_names = [alg_name for alg_name, alg_dict in self.results_dict.items()
                            if set(alg_dict.keys()) == set(self.task_names)]
        return self.filter_alg_names(common_alg_names)

    def analyze_errors(self):
        n_steps = 0
        for alg_name, task_results in self.results_dict.items():
            for task_name, split_results in task_results.items():
                for split_idx, split_result in enumerate(split_results):
                    for al_stat_idx, al_stat in enumerate(split_result['al_stats']):
                        n_steps += 1
                        if 'selection_status' in al_stat and al_stat['selection_status'] is not None:
                            print(f'Alg {alg_name} failed on step {al_stat_idx} of split {split_idx} of task {task_name}:',
                                  al_stat['selection_status'])

        print(f'Total number of DBAL steps across all experiments: {n_steps}')

    def analyze_eff_dims(self):
        n_larger = 0
        n_total = 0
        eff_dim_sum_grad = 0.0
        eff_dim_sum_ll = 0.0
        for alg_name in self.results_dict:
            if not alg_name.endswith('_grad_rp-512'):
                continue
            alg_name_ll = alg_name.replace('_grad_rp-512', '_ll')
            for task_name, split_results in self.results_dict[alg_name].items():
                if not task_name.endswith('256x16'):
                    continue
                for split_idx, split_result in enumerate(split_results):
                    for al_stat_idx, al_stat in enumerate(split_result['al_stats']):
                        eff_dim = al_stat['eff_dim']
                        al_stat_ll = self.results_dict[alg_name_ll][task_name][split_idx]['al_stats'][al_stat_idx]
                        eff_dim_ll = al_stat_ll['eff_dim']
                        n_total += 1
                        eff_dim_sum_grad += eff_dim
                        eff_dim_sum_ll += eff_dim_ll
                        if eff_dim > eff_dim_ll:
                            n_larger += 1

        print(f'eff dim was larger for grad_rp-512 than for ll in {100*n_larger/n_total:g}% of cases')
        print(f'avg eff dim for grad_rp-512: {eff_dim_sum_grad/n_total:g}')
        print(f'avg eff dim for ll: {eff_dim_sum_ll/n_total:g}')

    @staticmethod
    def load(exp_name: str) -> 'ExperimentResults':
        results_path = Path(custom_paths.get_results_path()) / exp_name
        pkl_filename = Path(custom_paths.get_cache_path()) / exp_name / 'results.pkl'
        results = None
        # first try to load from cached pkl file
        if utils.existsFile(pkl_filename) \
            and os.path.getmtime(pkl_filename) >= utils.last_mod_time_recursive(str(results_path)):
            try:
                results = utils.deserialize(pkl_filename)
            except Exception as e:
                print(f'Received exception while trying to load cached results, '
                      f'reloading results without cache. Exception: {e}')

        # if loading cached data did not work, load from scratch
        if results is None:
            results = {}
            for task_path in results_path.iterdir():
                task_name = task_path.name
                for alg_path in task_path.iterdir():
                    alg_name = alg_path.name
                    alg_results = {}
                    split_exists = False
                    for split_path in sorted(alg_path.iterdir(), key=lambda path: int(path.name)):
                        results_file = split_path / 'results.json'
                        if utils.existsFile(results_file):
                            split_results = utils.deserialize(split_path / 'results.json', use_json=True)
                            if task_name in alg_results:
                                alg_results[task_name].append(split_results)
                            else:
                                alg_results[task_name] = [split_results]
                            split_exists = True

                    if split_exists:
                        if alg_name in results:
                            results[alg_name].update(alg_results)
                        else:
                            results[alg_name] = alg_results
            utils.serialize(pkl_filename, results)
        return ExperimentResults(results_dict=results, exp_name=exp_name)


def get_latex_metric_name(metric_name: str) -> str:
    conversion_dict = {'mae': r'MAE',
                       'rmse': r'RMSE',
                       'maxe': r'MAXE',
                       'q95': r'95\% quantile',
                       'q99': r'99\% quantile'
                       }
    return conversion_dict[metric_name]


def get_latex_ds_name(ds_name: str) -> str:
    conversion_dict = {'ct': 'ct_slices',
                       'kegg_undir_uci': 'kegg_undir',
                       'query_agg_count': 'query',
                       'road_network': 'road',
                       'wecs': 'wec_sydney'}
    if ds_name in conversion_dict:
        ds_name = conversion_dict[ds_name]
    return ds_name.replace('_', r'\_')


def get_latex_task(task: str) -> str:
    conversion_dict = {'online_video': r'online\_video',
                       'sgemm': r'sgemm',
                       'kegg_undir_uci': r'kegg\_undir',
                       'stock': r'stock',
                       'wecs': r'wec\_sydney',
                       'sarcos': r'sarcos',
                       'diamonds': r'diamonds',
                       'fried': r'fried',
                       'road_network': r'road',
                       'poker': r'poker',
                       'mlr_knn_rng': r'mlr\_knn\_rng',
                       'methane': r'methane',
                       'protein': r'protein',
                       'ct': r'ct\_slices',
                       'query_agg_count': r'query'
                       }
    return conversion_dict[task]


def get_latex_selection_method(selection_method: str) -> str:
    conversion_dict = {'random': r'\textsc{Random}',
                       'fw': r'\textsc{FrankWolfe}',
                       'bait-f': r'\textsc{Bait-F}',
                       'bait-fb': r'\textsc{Bait-FB}',
                       'kmeanspp': r'\textsc{KMeansPP}',
                       'lcmd': r'\textsc{LCMD}',
                       'maxdet': r'\textsc{MaxDet}',
                       'maxdist': r'\textsc{MaxDist}',
                       'maxdiag': r'\textsc{MaxDiag}'
                       }

    parts = selection_method.split('-')
    method_name = '-'.join(parts[:-1]) if len(parts) > 1 else selection_method
    result = conversion_dict[method_name]
    if len(parts) > 1:
        result += '-TP' if parts[-1] == 'tp' else '-P'
    if method_name == 'lcmd':
        result += ' (ours)'
    return result


def get_latex_kernel(base_kernel: str, kernel_transformations: List[Tuple[str, List]], n_models: int) -> str:
    conversion_base_kernel_dict = {'grad': r'\mathrm{grad}',
                                   'll': r'\mathrm{ll}',
                                   'linear': r'\mathrm{lin}',
                                   'nngp': r'\mathrm{nngp}'}

    steps = [conversion_base_kernel_dict[base_kernel]]

    for name, args in kernel_transformations:
        if name == 'train':
            steps.append(r'\mathcal{X}_{\operatorname{train}}')
        elif name == 'scale':
            steps.append(r'\operatorname{scale}(\mathcal{X}_{\operatorname{train}})')
        elif name == 'rp':
            steps.append(r'\operatorname{rp}(' + str(args[0]) + ')')
        elif name == 'ens':
            steps.append(r'\operatorname{ens}(' + str(n_models) + ')')
        elif name == 'acs-rf':
            steps.append(r'\operatorname{acs-rf}(' + str(args[0]) + ')')
        elif name == 'acs-rf-hyper':
            steps.append(r'\operatorname{acs-rf-hyper}(' + str(args[0]) + ')')
        elif name == 'acs-grad':
            steps.append(r'\operatorname{acs-grad}')
        else:
            raise ValueError(f'Unknown kernel transformation "{name}"')

    return '$k_{' + r' \to '.join(steps) + '}$'


def save_latex_table_all_algs(results: ExperimentResults, filename: str):
    # creates a table for all algorithms, all metric are averaged over all data sets
    # Selection method | Kernel | MAE | RMSE | 95% | 99% | MAXE | avg. time

    ds_names = list({'_'.join(task_name.split('_')[:-1]) for task_name in results.task_names})
    all_alg_names = results.alg_names

    mae = results.get_avg_errors('mae')
    rmse = results.get_avg_errors('rmse')
    q95 = results.get_avg_errors('q95')
    q99 = results.get_avg_errors('q99')
    maxe = results.get_avg_errors('maxe')
    kernel_time = results.get_avg_al_times('kernel_time').select_split(9)
    selection_time = results.get_avg_al_times('selection_time').select_split(9)

    metrics_and_names = [('MAE', mae), ('RMSE', rmse), (r'95\%', q95), (r'99\%', q99), ('MAXE', maxe),
                         ('kernel_time', kernel_time), ('selection_time', selection_time)]
    alg_metrics = {alg_name: {name: np.mean([np.mean(metric.results_dict[alg_name][ds_name + '_256x16'])
                                             for ds_name in ds_names])
                    for name, metric in metrics_and_names} for alg_name in all_alg_names}

    n_digits = 3
    best_alg_names_per_metric = {}
    for metric_name, _ in metrics_and_names:
        rounded_results = [round(alg_metrics[alg_name][metric_name], n_digits) for alg_name in all_alg_names]
        min_result = np.min(rounded_results)
        best_alg_names_per_metric[metric_name] = [all_alg_names[i] for i in range(len(all_alg_names))
                                                  if rounded_results[i] == min_result]

    table_rows = {}
    table_header = '\\begin{tabular}{cccccccc}\n' + \
                   ' & '.join([r'Selection method', r'Kernel', r'MAE', r'RMSE', r'95\%', r'99\%',
                               r'MAXE', r'avg.\ time [$s$]']) + '\\\\\n\\hline\n'
    table_footer = '\n\\end{tabular}'

    # raw_sel_order = {'random': 0, 'maxdiag': 1, 'maxdet': 2, 'fw': 3, 'maxdist': 4, 'kmeanspp': 5, 'lcmd': 6}
    sel_name_order = ['random', 'maxdiag', 'maxdet', 'bait', 'fw', 'maxdist', 'kmeanspp', 'lcmd']

    raw_sel_names = {}

    for name_alg in all_alg_names:
        config = next(iter(results.results_dict[name_alg].values()))[0]['config']
        base_kernel = config['base_kernel']
        kernel_transformations = config['kernel_transforms']
        n_models = config.get('n_models', 1)

        raw_sel_name = name_alg.split('_')[1].split('-')[0]
        sel_name = get_latex_selection_method(name_alg.split('_')[1])
        raw_sel_names[name_alg] = raw_sel_name
        kernel_name = get_latex_kernel(base_kernel, kernel_transformations, n_models=n_models)
        if raw_sel_name == 'random':
            kernel_name = '---'
        alg_results = []
        for metric_name, _ in metrics_and_names[:5]:
            value_str = f'{alg_metrics[name_alg][metric_name]:5.3f}'
            if name_alg in best_alg_names_per_metric[metric_name]:
                value_str = r'\textbf{' + value_str + r'}'
            alg_results.append(value_str)
        alg_time = alg_metrics[name_alg]['kernel_time'] + alg_metrics[name_alg]['selection_time']

        row_strs = [sel_name, kernel_name] + alg_results + [f'{alg_time:5.3f}']
        table_rows[name_alg] = ' & '.join(row_strs)

    sub_groups = [[alg_name for alg_name in all_alg_names if raw_sel_names[alg_name] == sel_name] for sel_name in sel_name_order]
    sub_groups = [sub_group for sub_group in sub_groups if len(sub_group) > 0]
    for i in range(len(sub_groups)):
        sub_groups[i].sort(key=lambda alg_name: alg_metrics[alg_name]['RMSE'])

    sub_group_strs = ['\\\\\n'.join([table_rows[alg_name] for alg_name in sub_group]) for sub_group in sub_groups]

    result_str = table_header + ' \\\\\n\\hline\n'.join(sub_group_strs) + table_footer

    utils.writeToFile(Path(custom_paths.get_plots_path()) / results.exp_name / filename, result_str)


def save_latex_table_data_sets(results: ExperimentResults, filename: str, use_log: bool = True,
                               use_last_error: bool = False, metric_name: str = 'rmse'):
    ds_names = list({'_'.join(task_name.split('_')[:-1]) for task_name in results.task_names})
    ds_names.sort()
    alg_names = results.alg_names

    if use_last_error:
        rmse = results.get_last_errors(metric_name, use_log=use_log)
    else:
        rmse = results.get_avg_errors(metric_name, use_log=use_log)
    alg_metrics = {alg_name: {ds_name: np.mean(rmse.results_dict[alg_name][ds_name + '_256x16'])
                              for ds_name in ds_names} for alg_name in alg_names}

    n_digits = 3
    best_alg_names_per_data_set = {}
    for ds_name in ds_names:
        rounded_results = [round(alg_metrics[alg_name][ds_name], n_digits) for alg_name in alg_names]
        min_result = np.min(rounded_results)
        best_alg_names_per_data_set[ds_name] = [alg_names[i] for i in range(len(alg_names)) if
                                                rounded_results[i] == min_result]

    table_list = []
    table_header = '\\begin{tabular}{' + ('c' * (len(alg_names)+1)) + '}\n' + \
                   ' & '.join([r'Data set'] + [get_latex_selection_method(alg_name.split('_')[1])
                                               for alg_name in alg_names]) + '\\\\\n\\hline\n'
    table_footer = '\n\\end{tabular}'

    for ds_name in ds_names:
        ds_results = []
        for name_alg in alg_names:
            value_str = f'{alg_metrics[name_alg][ds_name]:5.3f}'
            if name_alg in best_alg_names_per_data_set[ds_name]:
                value_str = r'\textbf{' + value_str + r'}'
            ds_results.append(value_str)

        row_strs = [get_latex_task(ds_name)] + ds_results
        table_list.append(' & '.join(row_strs))

    table_list = table_list

    result_str = table_header + ' \\\\\n'.join(table_list) + table_footer

    utils.writeToFile(Path(custom_paths.get_plots_path()) / results.exp_name / filename, result_str)


def print_single_task_results(exp_results: ExperimentResults, task_name: str):
    """
    Prints results for a single task. For each log metric value, the function prints
    mean +- standard error (estimated standard deviation of the mean estimator,
    when considering the data sets to be fixed)
    :param exp_results: ExperimentResults object whose results should be printed.
    :param task_name: Task for which the results should be printed.
    """
    metric_names = ['mae', 'rmse', 'q95', 'q99', 'maxe']

    print(f'Results for task {task_name}:')
    exp_results = exp_results.filter_task_names([task_name]).filter_common_algs()
    alg_names = exp_results.alg_names
    for metric_name in metric_names:
        avg_results = exp_results.get_avg_errors(metric_name)
        results = avg_results.results_dict
        print(f'Results for metric {metric_name}:')
        alg_means = {alg_name: np.mean(results[alg_name][task_name]) for alg_name in alg_names}
        alg_stds = {alg_name: np.std(results[alg_name][task_name])/np.sqrt(len(results[alg_name][task_name]))
                    for alg_name in alg_names}

        alg_names_sorted = utils.dict_argsort(alg_means)
        str_table = []
        for alg_name in alg_names_sorted:
            str_table.append([alg_name + ': ', f'{alg_means[alg_name]:5.3f} ', '+- ', f'{alg_stds[alg_name]:5.3f}'])
        print(utils.pretty_table_str(str_table))
        print()

    print('\n\n')


def print_all_task_results(exp_results: ExperimentResults):
    for task_name in exp_results.task_names:
        print_single_task_results(exp_results, task_name)


def print_avg_results(exp_results: ExperimentResults, relative_to: Optional[str] = None,
                      filter_suffix: str = '256x16'):
    """
    Prints experiment results averaged over splits, steps and data sets.
    :param exp_results: ExperimentResults object whose results should be printed.
    :param relative_to: alg_name (or part thereof) of a method that results should be compared to.
    If None, the log metric values are printed directly +- the standard error where the data sets are considered fixed.
    Otherwise, the difference in log metric values between the respective method and the relative_to method is printed,
    +- the standard error thereof where the data sets are considered to be randomly drawn.
    In the latter case, the standard error is going to be larger
    but indicates how strongly the relative performance varies across data sets.
    :param filter_suffix: Only tasks with this suffix are going to be analyzed.
    """
    if relative_to is not None:
        print(f'Averaged results across tasks, relative to {relative_to}:')
    else:
        print('Averaged results across tasks:')

    exp_results = exp_results.filter_task_suffix(filter_suffix).filter_common_algs()

    metric_names = ['mae', 'rmse', 'q95', 'q99', 'maxe']

    # find all algorithms that occur in all tasks
    alg_names = exp_results.alg_names
    task_names = exp_results.task_names
    relative_to_name = None if relative_to is None else \
        (relative_to if relative_to in alg_names
        else [name for name in alg_names if relative_to in name][0])

    for metric_name in metric_names:
        print(f'Results for metric {metric_name}:')
        avg_results = exp_results.get_avg_errors(metric_name)
        alg_task_means = {alg_name: {task_name: np.mean(avg_results.results_dict[alg_name][task_name])
                                     for task_name in task_names}
                          for alg_name in alg_names}
        if relative_to is not None:
            alg_task_means = {alg_name: {task_name: alg_task_means[alg_name][task_name]
                                                    - alg_task_means[relative_to_name][task_name]
                                         for task_name in task_names}
                              for alg_name in alg_names}
        alg_means = {alg_name: np.mean(list(alg_task_means[alg_name].values())) for alg_name in alg_names}
        if relative_to is not None:
            # take std over tasks
            alg_stds = {alg_name: np.std(list(task_means.values())) / np.sqrt(len(list(task_means.values())))
                        for alg_name, task_means in alg_task_means.items()}
        else:
            # take sd over splits
            alg_stds = {alg_name: np.linalg.norm([np.std(avg_results.results_dict[alg_name][task_name])
                                                  / np.sqrt(len(avg_results.results_dict[alg_name][task_name]))
                                                  for task_name in task_names]) / len(task_names)
                        for alg_name in alg_names}

        alg_names_sorted = utils.dict_argsort(alg_means)
        str_table = []
        for alg_name in alg_names_sorted:
            str_table.append([alg_name + ': ', f'{alg_means[alg_name]:5.3f} ', '+- ', f'{alg_stds[alg_name]:5.3f}'])
        print(utils.pretty_table_str(str_table))
        print()

    print('\n\n')



import time
import dill
import torch.multiprocessing as mp
import os
import sys
import traceback

from .train import *


def get_cuda_devices():
    return [torch.device(f'cuda:{i}') for i in range(torch.cuda.device_count())]


def get_devices(use_gpu=True):
    return get_cuda_devices() if (use_gpu and torch.cuda.is_available()) else [torch.device('cpu')]


class GPUInfo:
    """
    Internal class.
    Provides some static methods that can be used to get information about (NVidia) GPUs.
    """
    _initialized = False

    @staticmethod
    def init():
        if not GPUInfo._initialized:
            import pynvml
            pynvml.nvmlInit()
            GPUInfo._initialized = True

    @staticmethod
    def get_process_ram_usage_gb(gpu_num: int, pid: int) -> float:
        # see https://github.com/gpuopenanalytics/pynvml/issues/21
        GPUInfo.init()
        import pynvml
        handle = pynvml.nvmlDeviceGetHandleByIndex(gpu_num)
        for proc in pynvml.nvmlDeviceGetComputeRunningProcesses(handle):
            if proc.pid == pid:
                return proc.usedGpuMemory / (1024. ** 3)
        else:
            return 0.0

    @staticmethod
    def get_total_ram_gb(gpu_num: int) -> float:
        GPUInfo.init()
        import pynvml
        h = pynvml.nvmlDeviceGetHandleByIndex(gpu_num)
        info = pynvml.nvmlDeviceGetMemoryInfo(h)
        total = info.total
        return total / (1024. ** 3)


class DeviceInfo:
    """
    Internal class.
    Provides some static methods that can be used to get information about available RAM and RAM usage on a device.
    """
    @staticmethod
    def get_total_ram_gb(device: str) -> float:
        device = str(device)
        if device == 'cpu':
            import psutil
            mem = psutil.virtual_memory()
            return mem.available / (1024. ** 3)
        elif device.startswith('cuda:'):
            gpu_num = int(device[5:])
            return GPUInfo.get_total_ram_gb(gpu_num=gpu_num)
        else:
            raise ValueError(f'Unknown device "{device}"')

    @staticmethod
    def get_process_ram_usage_gb(device: str, pid: Optional[int] = None) -> float:
        if pid is None:
            pid = os.getpid()  # get pid of this process
        device = str(device)
        if device == 'cpu':
            import psutil
            return psutil.Process(pid).memory_info().rss / 1024 ** 3
        elif device.startswith('cuda:'):
            gpu_num = int(device[5:])
            return GPUInfo.get_process_ram_usage_gb(gpu_num=gpu_num, pid=pid)
        else:
            raise ValueError(f'Unknown device "{device}"')


def measure_fixed_rams_gb(devices: List[str]) -> List[float]:
    """
    Internal method. Provides the RAM that is already used when using any PyTorch tensor on these devices.
    This can be significantly large due to the CUDA runtimes.
    :param devices: List of devices that should be used.
    :return: Returns a list of RAM usages (in GB) on these devices.
    """
    import torch
    # alloc dummy tensors to know how much memory PyTorch uses for its runtime
    dummy_tensors = [torch.ones(1).to(device) for device in devices]
    return [DeviceInfo.get_process_ram_usage_gb(device) for device in devices]


class FunctionRunner:
    """
    Simple helper class to run a function in a process / thread and submit the result back to a queue
    """
    def __init__(self, dill_f_and_args, result_queue):
        """
        :param dill_f_and_args: Should be a string that can be decoded with dill,
        i.e., dill_f_and_args = dill.dumps((f, args)) when we want to call f(*args) in the process.
        :param result_queue: Queue where the result should be pushed to.
        """
        self.dill_f_and_args = dill_f_and_args
        self.result_queue = result_queue

    def __call__(self):
        """
        Runs the function with the specified arguments and puts the result in the queue.
        """
        try:
            f, args = dill.loads(self.dill_f_and_args)
            result = f(*args)
            self.result_queue.put(result)
            self.result_queue.join()
        except Exception as e:
            print('Handling exception')
            print(e)
            traceback.print_exc()


class FunctionProcess:
    """
    Helper class to run a function in a separate process.
    """
    def __init__(self, f, *args):
        """
        :param f: Function to run in the process.
        :param args: Arguments to the function.
        """
        self.result_queue = mp.JoinableQueue()
        self.target = FunctionRunner(dill.dumps((f, args)), self.result_queue)
        self.process = mp.Process(target=self.target)

    def start(self) -> 'FunctionProcess':
        """
        Start the process computing the function.
        :return: Returns self such that we can use FunctionProcess(f, *args).start().
        """
        self.process.start()
        return self

    def is_done(self) -> bool:
        """
        :return: Returns true if the result of the function is available.
        """
        return not self.result_queue.empty()

    def get_ram_usage_gb(self, device='cpu') -> float:
        """
        :param device: PyTorch device string.
        :return: Returns the RAM usage in GB on the given device.
        """
        device = str(device)
        return DeviceInfo.get_process_ram_usage_gb(device, self.process.pid)

    def pop_result(self) -> Any:
        """
        :return: Returns the result and terminates the process, i.e., this function can only be called once.
        """
        result = self.result_queue.get()
        self.result_queue.task_done()
        self.process.terminate()
        self.process.join()
        return result


class AbstractJob:
    """
    Abstract base class for jobs that support execution and can return a RAM requirement and a name.
    """
    def __call__(self, device: str) -> None:
        """
        Abstract method to execute the job on the given device.
        :param device: PyTorch device to run the job on.
        """
        raise NotImplementedError()

    def get_ram_usage_gb(self) -> float:
        """
        Abstract method.
        :return: Should return the required RAM usage on the device in GB.
        Note that fixed RAM on the device due to PyTorch's CUDA runtime should not be included,
        as this is added automatically.
        """
        raise NotImplementedError()

    def get_desc(self) -> str:
        """
        Abstract method.
        :return: Should return a name that can be used for printing which job is running.
        """
        raise NotImplementedError()


class JobScheduler:
    """
    This class allows to run jobs with RAM constraints in separate processes on multiple devices in parallel.
    """
    _has_start_method_been_set = False

    def __init__(self, devices: Optional[List[str]] = None, use_gpu: bool = True, max_jobs_per_device: int = 1000):
        """
        :param devices: Optional list of PyTorch device strings.
        If None, the CPU is used if use_gpu=False and all available GPUs otherwise.
        :param use_gpu: Whether GPUs should be used if devices is None.
        :param max_jobs_per_device: Maximum number of jobs that are allowed to run per device.
        Note that the jobs also have RAM constraints,
        so the JobScheduler may run less jobs than the specified maximum number.
        """
        self.devices = (get_devices() if use_gpu else ['cpu']) if devices is None else devices
        self.max_jobs_per_device = max_jobs_per_device

    def run_all(self, jobs: List[AbstractJob]):
        """
        Run all jobs in separate processes, in parallel on multiple devices,
        according to the constraints imposed by the job's RAM bound and the maximum number of jobs per device.
        :param jobs: List of jobs.
        """
        if len(jobs) == 0:
            return
        if not JobScheduler._has_start_method_been_set:
            mp.set_start_method('spawn')
            JobScheduler._has_start_method_been_set = True
        start_time = time.time()

        print(f'Start time: {utils.format_date_s(start_time)}')

        max_ram_fraction = 0.9
        ram_gb_per_device = [DeviceInfo.get_total_ram_gb(device) * max_ram_fraction for device in self.devices]
        # execute this in a process such that reserved GPU memory can be released again
        fixed_ram_gb_per_device = FunctionProcess(measure_fixed_rams_gb, self.devices).start().pop_result()

        # started_infos: [{'process': ..., 'device': ..., 'ram_gb': ...}]
        started_infos = []
        # maybe start times?
        next_job_index = 0

        while next_job_index < len(jobs):
            next_job = jobs[next_job_index]
            job_ram_gb = next_job.get_ram_usage_gb()

            # try assigning a job to a device
            for i, device in enumerate(self.devices):
                n_jobs_on_device = len([si for si in started_infos if si['device'] == device])
                if n_jobs_on_device < self.max_jobs_per_device:
                    needs_ram_gb = job_ram_gb + fixed_ram_gb_per_device[i]
                    if needs_ram_gb > ram_gb_per_device[i]:
                        raise RuntimeError(f'RAM estimate of {needs_ram_gb:g} GB for job {next_job.get_desc()}'
                                           f' is too large for device')
                    remaining_ram = ram_gb_per_device[i] - sum([si['ram_gb'] for si in started_infos
                                                                if si['device'] == device], 0.0)
                    if needs_ram_gb < remaining_ram:
                        print(f'Starting job {next_job_index+1}/{len(jobs)} after {utils.format_length_s(time.time() - start_time)}')
                        started_infos.append({'process': FunctionProcess(next_job, device).start(),
                                              'device': device, 'ram_gb': needs_ram_gb})
                        next_job_index += 1
                        break
            else:  # no suitable device found
                # check which jobs are terminated
                new_terminated = False
                for i, si in enumerate(started_infos):
                    if si['process'].is_done():
                        si['process'].pop_result()
                        new_terminated = True
                        del started_infos[i]
                if not new_terminated:
                    time.sleep(0.1)

        # join all remaining processes
        for si in started_infos:
            si['process'].pop_result()

        end_time = time.time()
        print(f'End time: {utils.format_date_s(end_time)}')
        print(f'Total time: {utils.format_length_s(end_time - start_time)}')


class BatchALJob(AbstractJob):
    """
    This class implements the type of jobs used in Batch Active Learning, such that they can be used in JobScheduler.
    """
    def __init__(self, task: Task, split_id: int, trainer: ModelTrainer, ram_gb_per_sample: float, exp_name: str,
                 use_pool_for_normalization: bool, do_timing: bool, ram_gb_per_sample_bs: float):
        """
        :param task: Task to run.
        :param split_id: ID of the task split to run.
        :param trainer: ModelTrainer used for Batch Active Learning.
        :param ram_gb_per_sample: This should be an upper bound on how much RAM (in GB)
        per sample of the train+pool sets will be used. This might be on the order of 1e-5 or so.
        For details, see the implementation of get_ram_usage_gb().
        """
        self.task = task
        self.split_id = split_id
        self.trainer = trainer
        self.ram_gb_per_sample = ram_gb_per_sample
        self.ram_gb_per_sample_bs = ram_gb_per_sample_bs
        self.exp_name = exp_name
        self.use_pool_for_normalization = use_pool_for_normalization
        self.do_timing = do_timing

    def __call__(self, device: str):
        try:
            result_dict = self.trainer(TaskSplit(self.task, id=self.split_id,
                                                 use_pool_for_normalization=self.use_pool_for_normalization),
                                       device=device, do_timing=self.do_timing)
            utils.serialize(self.trainer.get_result_file_path(self.exp_name, self.task.task_name, self.split_id), result_dict,
                            use_json=True)
        except Exception as e:
            print(f'Error for alg {self.trainer.alg_name} on split {self.split_id} of task {self.task.task_name}:',
                  file=sys.stderr)
            print(e, file=sys.stderr)
            traceback.print_exc()
            print('', file=sys.stderr, flush=True)

    def get_ram_usage_gb(self) -> float:
        max_bs = (max(self.task.al_batch_sizes) if len(self.task.al_batch_sizes) > 0 else 0)
        return 0.2 + self.task.data_info.n_samples * self.task.data_info.n_features * 8 / (1024**3) \
               + (self.task.n_train + self.task.n_pool) * self.ram_gb_per_sample \
               + (self.task.n_train + self.task.n_pool) * max_bs * self.ram_gb_per_sample_bs

    def get_desc(self) -> str:
        return f'{self.trainer.alg_name} on split {self.split_id} of task {self.task.task_name}'


class JobRunner:
    """
    Simple class for adding Batch AL jobs and then running them in parallel using a JobScheduler.
    """
    def __init__(self, scheduler: JobScheduler):
        """
        :param scheduler: The JobScheduler configuration that should be used to run the jobs.
        """
        self.scheduler = scheduler
        self.jobs = []

    def add(self, exp_name: str, split_id: int, tasks: List[Task], ram_gb_per_sample: float,
            trainer: ModelTrainer, do_timing: bool, warn_if_exists: bool = True,
            use_pool_for_normalization: bool = True, ram_gb_per_sample_bs: float = 0.0):
        """
        Adds jobs for each task in tasks and for each split in range(self.n_splits)
        to the list of jobs that will be run in self.run_all().
        :param exp_name: experiment group name, is used for the top-level folder for saving
        :param split_id: id of the split to run on
        :param tasks: List of tasks that the trainer will be run on.
        :param ram_gb_per_sample: RAM estimate (in GB) per sample, typically of the order of 1e-5.
        An upper bound on the RAM usage (in GB) of a process is computed as
        0.2 + (size of data set in GB) + ram_gb_per_sample * (n_train + n_pool).
        :param trainer: ModelTrainer that runs a certain configuration of batch active learning.
        :param do_timing: Whether to take extra efforts (CUDA synchronization) for proper timing.
        :param warn_if_exists: Whether to print a message if the results already exist
        :param use_pool_for_normalization: whether to compute data normalization statistics also based on the pool data
        """
        for task in tasks:
            if utils.existsFile(trainer.get_result_file_path(exp_name, task.task_name, split_id)):
                if warn_if_exists:
                    print(f'Results already exist for {trainer.alg_name} on split {split_id} of task {task.task_name}',
                          flush=True)
                continue
            self.jobs.append(BatchALJob(task, split_id, trainer, ram_gb_per_sample, exp_name=exp_name,
                                        use_pool_for_normalization=use_pool_for_normalization,
                                        do_timing=do_timing, ram_gb_per_sample_bs=ram_gb_per_sample_bs))

    def run_all(self):
        """
        Runs all jobs on the job scheduler.
        """
        self.scheduler.run_all(self.jobs)







from setuptools import setup, find_packages

# following https://betterscientificsoftware.github.io/python-for-hpc/tutorials/python-pypi-packaging/
# and https://packaging.python.org/en/latest/guides/making-a-pypi-friendly-readme/

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='bmdal_reg',
    version='2.1.4',    
    description='Deep Batch Active Learning for Regression',
    url='https://github.com/dholzmueller/bmdal_reg',
    author='David Holzmüller',
    author_email='nextuser@live.de',
    license='Apache 2.0',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=['torch',
                      'numpy',
                      'dill',
                      'psutil',
                      'matplotlib',
                      'seaborn',
                      'pandas',
                      'openml',
                      'mat4py'                     
                      ],

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',  
        'Operating System :: POSIX :: Linux',        
        'Programming Language :: Python :: 3',
    ],
)




from pathlib import Path
import shutil
from typing import *

from . import custom_paths
from . import utils

# This file contains some utility functions to modify/rename/remove saved results.
# It can be used for example if the names of some experiment results should be changed.


def rename_alg(exp_name: str, old_name: str, new_name: str):
    print(f'Renaming alg "{old_name}" to "{new_name}" for {exp_name} experiments')
    results_path = Path(custom_paths.get_results_path()) / exp_name
    for task_path in results_path.iterdir():
        if utils.existsDir(task_path / old_name):
            shutil.move(task_path / old_name, task_path / new_name)


def remove_alg(exp_name: str, alg_name: str):
    print(f'Removing alg "{alg_name}" for {exp_name} experiments')
    results_path = Path(custom_paths.get_results_path()) / exp_name
    for task_path in results_path.iterdir():
        if utils.existsDir(task_path / alg_name):
            shutil.rmtree(task_path / alg_name)


def replace_in_alg_name(exp_name: str, old_name: str, new_name: str):
    print(f'Replacing "{old_name}" with "{new_name}" in alg names for {exp_name} experiments')
    results_path = Path(custom_paths.get_results_path()) / exp_name
    for task_path in results_path.iterdir():
        for alg_path in task_path.iterdir():
            alg_name = str(alg_path.name)
            new_alg_name = alg_name.replace(old_name, new_name)
            if alg_name != new_alg_name:
                shutil.move(task_path / alg_name, task_path / new_alg_name)


def process_results(exp_name: str, f: Callable):
    print('Applying function to results for {exp_name} experiments')
    results_path = Path(custom_paths.get_results_path()) / exp_name
    for task_path in results_path.iterdir():
        for alg_path in task_path.iterdir():
            for split_path in alg_path.iterdir():
                file_path = split_path / 'results.json'
                if utils.existsFile(file_path):
                    results = utils.deserialize(file_path, use_json=True)
                    results = f(results)
                    utils.serialize(file_path, results, use_json=True)


if __name__ == '__main__':
    pass





from .train import *
from .data import *
from .task_execution import get_devices
import time


def run_single_task():
    """
    Test function for running Batch Active Learning on a single task.
    """
    task = Task.get_tabular_tasks(n_train=256, al_batch_sizes=[256] * 16, ds_names=['kegg_undir_uci'])[0]
    task_split = TaskSplit(task, id=0)
    sigma = 1e-3
    compute_eff_dim = True
    kwargs = dict(maxdet_sigma=sigma, bait_sigma=sigma, compute_eff_dim=compute_eff_dim, allow_float64=True,
                  lr=0.375, weight_gain=0.2, bias_gain=0.2, lr_sched='lin'
                  )
    trainer = ModelTrainer('NN_test', act='relu',
                           base_kernel='grad',
                           kernel_transforms=[
                               # ('scale', [None]),
                               ('rp', [2048]),
                               # ('acs-rf-hyper', [512, None]),
                               # ('acs-rf', [512]),
                               ('train', [sigma, None]),
                           ], selection_method='bait',
                           sel_with_train=True,
                           allow_maxdet_fs=False,
                           # print_effective_dimension=True,
                           n_epochs=256, n_models=1, al_on_cpu=False, **kwargs)
    # trainer = ModelTrainer('NN_test', act='relu',
    #                        base_kernel='ll',
    #                        kernel_transforms=[
    #                            # ('scale', [None]),
    #                            ('rp', [512]),
    #                            # ('acs-rf-hyper', [512, None]),
    #                            # ('acs-rf', [512]),
    #                            ('train', [sigma, None]),
    #                        ], selection_method='bait', overselection_factor=2.0,
    #                        sel_with_train=False,
    #                        allow_maxdet_fs=False,
    #                        # print_effective_dimension=True,
    #                        n_epochs=256, n_models=1, al_on_cpu=False, **kwargs)
    start_time = time.time()
    results = trainer(task_split, device=get_devices()[0], do_timing=False)
    print(results)
    print(f'Time: {time.time() - start_time:g}')


if __name__ == '__main__':
    run_single_task()
    # import cProfile
    # cProfile.run('run_single_task()')
    pass



from .feature_maps import *


class Features:
    """
    This class represents a combination of a feature map and feature data.
    Hence, it implicitly represents a (precomputed or non-precomputed) feature matrix.
    Whenever operations between to Features objects are used,
    it is assumed that both objects share the same feature map.
    Features objects can be transformed in various ways. Here is an example of a typical use-case:
    # initialize feature_map, train_feature_data, pool_feature_data
    train_features = Features(feature_map, train_feature_data)
    pool_features = Features(feature_map, pool_feature_data)
    tfm = train_features.posterior_tfm(sigma=0.1)
    post_train_features = tfm(train_features)
    post_pool_features = tfm(pool_features)
    post_pool_variances = post_pool_features.get_kernel_matrix_diag()
    """
    def __init__(self, feature_map: FeatureMap, feature_data: FeatureData, diag: Optional[torch.Tensor] = None):
        """
        :param feature_map: Feature map.
        :param feature_data: Data that serves as input to the feature map.
        :param diag: Optional parameter representing the precomputed kernel matrix diagonal, i.e.,
        feature_map.get_kernel_matrix_diag(feature_data).
        """
        self.feature_map = feature_map
        self.feature_data = feature_data
        self.diag = diag
        if diag is not None and not isinstance(diag, torch.Tensor):
            raise ValueError(f'diag has wrong type {type(diag)}')

    def precompute(self) -> 'Features':
        """
        :return: Returns a Features object where the feature map is precomputed on the feature data,
        i.e., some methods should be faster to evaluate on the precomputed Features object.
        """
        fm, fd = self.feature_map.precompute(self.feature_data)
        if self.diag is None:
            self.diag = fm.get_kernel_matrix_diag(fd)
        return Features(fm, fd, self.diag)

    def simplify(self) -> 'Features':
        """
        :return: Returns a Features object where the feature data is simplified (un-batched etc.),
        which potentially makes evaluations faster, similar to precompute().
        """
        return Features(self.feature_map, self.feature_data.simplify(), self.diag)

    def scale_tfm(self, factor: Optional[float] = None) -> 'FeaturesTransform':
        """
        :param factor: Factor by which to scale. If this is None,
        the factor is computed as the automatic scale normalization factor from the paper,
        on self.feature_data.
        :return: Returns a transformation object that scales a Features object by the given factor.
        """
        if factor is None:
            diag = self.get_kernel_matrix_diag()
            factor = 1.0 / math.sqrt(diag.mean().item())
        return LambdaFeaturesTransform(lambda f: Features(ScaledFeatureMap(f.feature_map, factor), f.feature_data))

    def posterior_tfm(self, sigma: float = 1.0, allow_kernel_space_posterior: bool = True, **config) \
            -> 'FeaturesTransform':
        """
        Computes the posterior transformation after observing self.feature_data.
        :param sigma: GP noise standard deviation.
        :param allow_kernel_space_posterior: Whether the method is allowed to use the kernel-space posterior formula
        if it is deemed more efficient but not strictly necessary. The kernel-space posterior feature map returned
        will not allow a computation of the feature matrix, which could be detrimental
        for methods that want to operate in feature space.
        :return: Returns a transformation object that replaces the feature map of a Features object
        by the posterior feature map arising from the GP with prior given by self.feature_map,
        after observing self.feature_data.
        """
        fm = self.feature_map.posterior(self.feature_data, sigma,
                                        allow_kernel_space_posterior=allow_kernel_space_posterior)
        return LambdaFeaturesTransform(lambda f, fm=fm: Features(fm, f.feature_data))

    def sketch_tfm(self, n_features: int, **config) -> 'FeaturesTransform':
        """
        Computes a sketching transformation for the current feature map.
        :param n_features: Number of target features of the sketched feature map.
        :param config: Optional parameters to specify sketching options.
        :return: Returns a Transformations that replaces the feature map of a Features object
        by the sketched version of self.feature_map.
        """
        fm = self.feature_map.sketch(n_features, **config)
        return LambdaFeaturesTransform(lambda f, fm=fm: Features(fm, f.feature_data))

    def acs_grad_tfm(self, sigma: float = 1.0) -> 'FeaturesTransform':
        """
        Computes the acs-grad transformation from the paper,
        using self.feature_data as training data to condition on.
        :param sigma: The GP noise standard deviation.
        :return: Returns a transformation object applying the acs-grad transformation.
        """
        post_features = self.posterior_tfm(sigma)(self)
        fm = ProductFeatureMap([self.feature_map, post_features.feature_map])
        if sigma != 1.0:
            fm = ScaledFeatureMap(fm, factor=1.0/sigma**4)
        return LambdaFeaturesTransform(lambda f, fm=fm: Features(fm, ListFeatureData([f.feature_data, f.feature_data])))

    def acs_rf_hyper_tfm(self, y_train: torch.Tensor, n_features: int, **config) \
            -> 'FeaturesTransform':
        """
        Computes the acs-rf-hyper transformation from the paper,
        using self.feature_data as training data to condition on.
        :param y_train: Labels corresponding to self.feature_data.
        Usually, self.feature_data corresponds to the training data, hence the name y_train.
        :param n_features: Number of target random features.
        :param config: Potential additional options (unused).
        :return: Returns a transformation that applies the acs-rf-hyper transformation to a Features object.
        """
        # adapted from https://github.com/rpinsler/active-bayesian-coresets
        x_train = self.get_feature_matrix()
        y_train = y_train.squeeze(1).type(x_train.type())
        in_features = x_train.shape[1]
        s = 1.0    # prior variance / length scale, I guess?
        # inverse gamma prior parameters
        a0 = 1.0
        b0 = 1.0
        y_var = b0 / a0
        w_cov_prior = s * torch.eye(in_features, device=x_train.device, dtype=x_train.dtype)
        xxw = x_train.t().matmul(x_train) + w_cov_prior
        L = robust_cholesky(xxw)  # xxw = L * L^T
        L_inv = torch.inverse(L)  # xxw^{-1} = L^{-T} * L^{-1}
        # theta_cov = torch.inverse(xxw)
        theta_cov = L_inv.t().matmul(L_inv)
        theta_mean = theta_cov.matmul(x_train.t().matmul(y_train))
        sigma_tilde_inv = xxw
        a_tilde = a0 + 0.5 * x_train.shape[0]
        b_tilde = b0 + 0.5 * ((y_train**2).sum() - torch.dot(theta_mean.t(), sigma_tilde_inv.matmul(theta_mean))).item()
        # theta_samples is be of shape in_features x n_features
        # if x \sim N(0, I), then L^{-T}x \sim N(0, L^{-T}L^{-1}) = N(0, theta_cov)
        theta_samples = L_inv.t().matmul(torch.randn(in_features, n_features, device=x_train.device, dtype=x_train.dtype))

        acs_rf_tfm = ACSRFHyperDataTransform(theta_cov=theta_cov, theta_samples=theta_samples, a_tilde=a_tilde,
                                             b_tilde=b_tilde, y_var=y_var)
        fm = SequentialFeatureMap(IdentityFeatureMap(n_features=n_features), [self.feature_map, acs_rf_tfm])
        return LambdaFeaturesTransform(lambda f, fm=fm: Features(fm, f.feature_data))

    def acs_rf_tfm(self, n_features: int, sigma: float = 1.0,
                      **config) -> 'FeaturesTransform':
        """
        Computes the acs-rf transformation from the paper,
        using self.feature_data as training data to condition on.
        :param n_features: Number of target random features.
        :param sigma: GP noise standard deviation.
        :param config: Potential additional options (unused).
        :return: Returns a transformation object applying the acs-rf transformation to a Features object.
        """
        # partially adapted from https://github.com/rpinsler/active-bayesian-coresets
        x_train = self.get_feature_matrix()
        in_features = x_train.shape[1]
        w_cov_prior = sigma**2 * torch.eye(in_features, device=x_train.device, dtype=x_train.dtype)
        xxw = x_train.t().matmul(x_train) + w_cov_prior
        L = robust_cholesky(xxw)  # xxw = L * L^T
        L_inv = torch.inverse(L)  # xxw^{-1} = L^{-T} * L^{-1}
        # theta_cov = torch.inverse(xxw)
        theta_cov = L_inv.t().matmul(L_inv)
        # theta_samples is be of shape in_features x n_features
        # if x \sim N(0, I), then L^{-T}x \sim N(0, L^{-T}L^{-1}) = N(0, theta_cov)
        theta_samples = L_inv.t().matmul(torch.randn(in_features, n_features, device=x_train.device, dtype=x_train.dtype))

        acs_rf_tfm = ACSRFDataTransform(theta_cov=theta_cov, theta_samples=theta_samples, sigma=sigma)
        fm = SequentialFeatureMap(IdentityFeatureMap(n_features=n_features), [self.feature_map, acs_rf_tfm])
        return LambdaFeaturesTransform(lambda f, fm=fm: Features(fm, f.feature_data))

    def get_n_samples(self) -> int:
        """
        :return: Returns the number of samples in self.feature_data.
        """
        return self.feature_data.get_n_samples()

    def __len__(self) -> int:
        """
        :return: Returns the number of samples in self.feature_data.
        """
        return self.get_n_samples()

    def get_n_features(self) -> int:
        """
        :return: Returns the number of features of the corresponding feature map.
        """
        return self.feature_map.get_n_features()

    def get_device(self) -> str:
        """
        :return: Returns the (torch) device that the feature data is on.
        """
        return self.feature_data.get_device()

    def get_dtype(self) -> Any:
        """
        :return: Returns the (torch) dtype that the feature data has.
        """
        return self.feature_data.get_dtype()

    def __getitem__(self, idxs: Union[int, slice, torch.Tensor]) -> 'Features':
        """
        Returns a Features object where the feature data is indexed by idxs. Note that if idxs is an integer,
        the feature data will be indexed by [idxs:idxs+1], i.e. the batch dimension will not be removed
        and the resulting feature data tensors will have shape [1, ...].
        This method is called when using an indexing expression
        such as features[0] or features[-2:] or features[torch.Tensor([0, 2, 4], dtype=torch.long)]
        :param idxs: Integer, slice, or torch.Tensor of integer type.
        See the comment above about indexing with integers.
        :return: Returns a Feature object where the feature data is the indexed version of self.feature_data.
        """
        idxs = Indexes(self.get_n_samples(), idxs)
        return Features(self.feature_map, self.feature_data[idxs],
                        None if self.diag is None else self.diag[idxs.get_idxs()])

    def get_kernel_matrix_diag(self) -> torch.Tensor:
        """
        Returns the kernel matrix diagonal
        obtained by self.feature_map.get_kernel_matrix_diag(self.feature_data).
        The kernel matrix diagonal is stored and reused,
        such that multiple calls to this method do not trigger multiple computations.
        Consequently, the returned Tensor should not be modified.
        :return: Returns a torch.Tensor of shape [len(self)] containing the kernel matrix diagonal.
        """
        if self.diag is None:
            self.diag = self.feature_map.get_kernel_matrix_diag(self.feature_data)
        return self.diag

    def get_kernel_matrix(self, other_features: 'Features') -> torch.Tensor:
        """
        Returns the kernel matrix k(self.feature_data, other_features.feature_data),
        where k is the kernel given by self.feature_map.
        :param other_features: Other features corresponding to the columns of the resulting kernel matrix.
        :return: Returns a torch.Tensor of shape [len(self), len(other_features)] corresponding to the kernel matrix.
        """
        return self.feature_map.get_kernel_matrix(self.feature_data, other_features.feature_data)

    def get_feature_matrix(self) -> torch.Tensor:
        """
        :return: Returns self.feature_map.get_feature_matrix(self.feature_data), i.e.,
        a torch.Tensor of shape [len(self), self.get_n_features()] containing the feature matrix.
        Note that this method cannot be used for all feature maps,
        since they might have an infinite-dimensional feature space.
        """
        return self.feature_map.get_feature_matrix(self.feature_data)

    def get_sq_dists(self, other_features: 'Features') -> torch.Tensor:
        """
        Return a matrix containing the squared feature space distances
        between self.feature_data and other_features.feature_data. These are computed using the kernel,
        hence they also work for feature maps with infinite-dimensional feature space.
        :param other_features: Features object to compute the distances to.
        :return: Returns a torch.Tensor of shape [len(self), len(other_features)]
        representing the squared distances between the feature data in feature space.
        """
        diag = self.get_kernel_matrix_diag()
        other_diag = other_features.get_kernel_matrix_diag()
        kernel_matrix = self.get_kernel_matrix(other_features)
        sq_dists = diag[:, None] + other_diag[None, :] - 2*kernel_matrix
        return sq_dists

    def batched(self, batch_size: int) -> 'Features':
        """
        Return a Features object that behaves as self,
        but where the feature data is virtually batched such that certain transformations are applied in batches.
        :param batch_size: Batch size of the batches. The last batch may be smaller.
        :return: Returns a Features object with batched feature data.
        """
        return Features(self.feature_map, self.feature_data.batched(batch_size), self.diag)

    def concat_with(self, other_features: 'Features'):
        """
        Concatenates two features objects along the sample dimension.
        :param other_features: Other Features object to concatenate with self.
        :return: Returns a Features object where self.feature_data and other_features.feature_data
        are concatenated along the batch dimension.
        """
        diag = torch.cat([self.diag, other_features.diag], dim=0) \
            if self.diag is not None and other_features.diag is not None \
            else None
        return Features(self.feature_map, ConcatFeatureData([self.feature_data, other_features.feature_data]), diag)


class ACSRFHyperDataTransform(DataTransform):
    """
    Internal helper class for the acs-rf-hyper transformation.
    This is essentially a feature map, transforming the data into random features.
    """
    def __init__(self, theta_cov: torch.Tensor, theta_samples: torch.Tensor, a_tilde: float, b_tilde: float,
                 y_var: float):
        super().__init__()
        self.theta_cov = theta_cov
        self.theta_samples = theta_samples
        self.a_tilde = a_tilde
        self.b_tilde = b_tilde
        self.y_var = y_var

    def forward(self, feature_data: FeatureData, idxs: Indexes) -> FeatureData:
        x = feature_data.get_tensor(idxs)
        n_features = x.shape[1]
        pred_var = self.b_tilde / self.a_tilde * (1 + torch.sum(x @ self.theta_cov * x, dim=-1))
        const = -0.5 * math.log(2 * math.pi * self.y_var)
        # z = (x_pool @ theta_sample)[:, None]
        z = x.matmul(self.theta_samples)
        # expected_ll = const - 0.5 / y_var * (z ** 2 - 2 * pred_mean * z + pred_var + pred_mean ** 2)
        expected_ll = const - 0.5 / self.y_var * (z ** 2 + pred_var[:, None])
        return TensorFeatureData(math.sqrt(1 / n_features) * expected_ll)


class ACSRFDataTransform(DataTransform):
    """
    Internal helper class for the acs-rf transformation.
    This is essentially a feature map, transforming the data into random features.
    """
    def __init__(self, theta_cov: torch.Tensor, theta_samples: torch.Tensor, sigma: float):
        super().__init__()
        self.theta_cov = theta_cov
        self.theta_samples = theta_samples
        self.sigma = sigma

    def forward(self, feature_data: FeatureData, idxs: Indexes) -> FeatureData:
        x = feature_data.get_tensor(idxs)
        n_features = x.shape[1]
        k_train = torch.sum(x @ self.theta_cov * x, dim=-1)[:, None]
        const = 0.5 * torch.log(1 + k_train / self.sigma**2)
        # z = (x_pool @ theta_sample)[:, None]
        z = x.matmul(self.theta_samples)
        # expected_ll = const - 0.5 / y_var * (z ** 2 - 2 * pred_mean * z + pred_var + pred_mean ** 2)
        expected_ll = const - 0.5 / self.sigma**2 * (z ** 2 + k_train)
        return TensorFeatureData(math.sqrt(1 / n_features) * expected_ll)


class FeaturesTransform:
    """
    Abstract base class for classes that allow to transform a Features object into another Features object
    """
    def __call__(self, features: Features) -> Features:
        """
        This method should be overridden by subclasses.
        :param features: Features object to transform.
        :return: Returns the transformed Features object.
        """
        raise NotImplementedError()


class LambdaFeaturesTransform(FeaturesTransform):
    """
    FeaturesTransform subclass that simply applies a Callable object to the Features.
    """
    def __init__(self, f: Callable[[Features], Features]):
        """
        :param f: Function to apply to the Features object.
        """
        self.f = f

    def __call__(self, features: Features) -> Features:
        return self.f(features)


class SequentialFeaturesTransform(FeaturesTransform):
    """
    FeaturesTransform subclass that operates by applying multiple FeaturesTransform objects.
    """
    def __init__(self, tfms: List[FeaturesTransform]):
        self.tfms = tfms

    def __call__(self, features: Features) -> Features:
        for tfm in self.tfms:
            features = tfm(features)
        return features


class PrecomputeTransform(FeaturesTransform):
    """
    Transformation that precomputes Features, possibly with batching.
    """
    def __init__(self, batch_size: int = -1):
        """
        :param batch_size: Batch size to apply to the precomputation. Set to -1 if batching should not be used.
        """
        self.batch_size = batch_size

    def __call__(self, features: Features) -> Features:
        if self.batch_size > 0:
            features = features.batched(self.batch_size)
        return features.precompute().simplify()


class BatchTransform(FeaturesTransform):
    """
    Transformation that batches Features.
    """
    def __init__(self, batch_size: int):
        """
        :param batch_size: Batch size to apply to the Features object.
        """
        self.batch_size = batch_size

    def __call__(self, features: Features) -> Features:
        return features.batched(self.batch_size)


